from ..exceptions import *


class Put():
    def __init__(self, requester):
        self.requester = requester
        
#SESSION

    def session(self):
        """ this updates session
        """
        endpoint = "/auth/token"
        
        return self.requester.put(endpoint)
        
#ACCOUNT AUTHORIZATION CODES
#ADMINISTRATORS
#ADVICE OF CHARGE
#ALTERNATE NUMBERS
#ANSWER CONFIRMATION
#ALTERNATE USER ID
#ANNOUNCEMENTS
#ANONYMOUS CALL REJECTION
#ATTENDANT CONSOLE
#AUTHENTICATION

    def user_authentication_service(self, user_id: str, new_password: str):
        """Set new SIP Authentication password for a single user.

        Args:
            user_id (str): Target user ID to reset the SIP authentication password.
            new_password (str): New SIP authentication password to apply to new user.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/users/authentication"
        
        data = {
            "userId": user_id,
            "newPassword": new_password
        }
        
        return self.requester.put(endpoint, data=data)
    
    
    def user_web_authentication_password(self, user_id: str, new_password: str):
        """Set new Web Authentication password for a single user.

        Args:
            user_id (str): Target user ID to reset the web authentication password.
            new_password (str): New web authentication password to apply to new user.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/users/passwords"
        
        data = {
            "userId": user_id,
            "newPassword": new_password
        }
        
        return self.requester.put(endpoint, data=data)

#AUTO ATTENDANTS

    def auto_attendants_status(self, auto_attendant_user_ids: list, 
                               status: bool =True):
        """Updates a list of auto attendants (AA) status to either active or inactive.

        Args:
            auto_attendant_user_ids (list): List of service user IDs (AA IDs), the status given 
            will be applied to these.
            status (bool, optional): Boolean value of True (Activate) or False (Deactivate) 
            which will be applied to list of AAs. Defaults to True.
            
        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/auto-attendants/status"
        
        data = {     
            "instances": [{'serviceUserId': auto_attendant_user_id, 'isActive': status} 
                          for auto_attendant_user_id in auto_attendant_user_ids]
        }
        
        return self.requester.put(endpoint, data=data)

    def auto_attendant(self, service_provider_id: str, group_id, auto_attendant_user_id: str, updates: dict):
        """Updates a single target Auto Attendant.
        
        Note: Needs the service instance profile to use this method.

        Args:
            service_provider_id (str): Service Provider ID where Group is hosted.
            group_id (str): Group ID where target Auto Attendant is located.
            auto_attendant_user_id (str): Target Auto Attendant User ID.
            updates (dict): Updates to be applied to Auto Attendant.

        Returns:
            Dict: Updated version of the Auto Attendant with updates applied. 
        """
        
        endpoint = "/groups/auto-attendants"
        
        if 'serviceInstanceProfile' not in updates:     
            updates.setdefault('serviceInstanceProfile', {})

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["serviceUserId"] = auto_attendant_user_id
    
        return self.requester.put(endpoint, data=updates)

    
    def auto_attendant_submenu(self, auto_attendant_user_id: str, 
                               submenu_id: str, updates: dict):
        """This method allows you to update the configuration of the submenus for your AAs
        
        Args:
            auto_attendant_user_id (str): Service user ID of your auto attendant.
            submenu_id (str): Service user ID of your submenu
            updates (dict): Updates your applying to your submenu.
            
        Returns:
            None: This method does not return any specific value.
        """
        
        
        endpoint = "/groups/auto-attendants/submenus"
        
        updates["serviceUserId"] = auto_attendant_user_id,
        updates["submenuId"] = submenu_id
        
        return self.request.put(endpoint, data=updates)
        

#AUTOMATIC CALLBACK
#AUTOMATIC HOLD RETRIEVE
#BARGE IN EXEMPT
#BASIC CALL LOGS
#BROADWORKS ANYWHERE
#BROADWORKS MOBILITY
#BROADWORKS NAVIGATION
#BROADWORKS RECEPTIONIST ENTERPRISE
#BROADWORKS RECEPTIONIST OFFICE
#BROADWORKS RECEPTIONIST SMALL BUSINESS
#BUSY LAMP FIELD
#CALL CAPACITY
#CALL CENTER

    def group_call_centers_status(self, call_center_user_ids: list, 
                                  status: bool =True):
        """Updates a list of call centers (CC) status to either active or inactive.

        Args:
            call_center_user_ids(list): List of service user IDs (CC IDs), the status given will be applied to these.
            status (bool): Boolean value of True (Activate) or False (Deactivate) which will be applied to list of AAs.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/call-centers/status"
        
        data = {     
            "instances": [{'serviceUserId': call_center_user_id, 'isActive': status} 
                          for call_center_user_id in call_center_user_ids]
        }
        
        return self.requester.put(endpoint, data=data)
        
        
    def group_call_center(self, call_center_user_id: str, updates: dict):
        """Allows you to update a specific call center.

        Args:
            call_center_user_id (str): Service user id of the target call center. 
            updates (dict): Updates to apply in in a dictionary format.

        Returns:
            Dict: The call center with the new applied updates.
        """
        
        endpoint = "/groups/call-centers"
        
        updates["serviceUserId"] = call_center_user_id

        return self.requester.put(endpoint, data=updates)
    
    
    def group_call_center_agents(self, call_center_user_id: str, 
                                 agent_user_ids: list):
        """Add or remove agents to a call center. 
        
        Note: Leave the agent_user_ids blank to remove all users and to remove some only include
        the users you would like to include in this call center.
        
        Args:
            call_center_user_id (str): Service user ID of the target call center.
            agent_user_ids (list): List of user IDs to be added to call center.

        Returns:
            Dict: Dictionary of the new state of the CC.
        """
        
        endpoint = "/groups/call-centers/agents"
        
        data = {
            "serviceUserId": call_center_user_id,
            "agents": [{"userId": agent_id} for agent_id in agent_user_ids]
        }

        return self.requester.put(endpoint, data=data)
        
    
    def group_call_center_agents_levels(self, call_center_user_id: str, 
                                        agent_user_ids: list, 
                                        skill_level: int):
        """Update a list of agents skill level in a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the call center users belong to. 
            agent_user_ids (list): List of the target users.
            skill_level (int): Skill level that will be applied to the list of users in the target call center.

        Returns:
            Dict: CC ID and list of the agent and their updated skill level.
        """
        
        endpoint = "/groups/call-centers/agents"
        
        data = {
            "serviceUserId": call_center_user_id,
            "agents": [{"userId": agent_id, "skillLevel": skill_level} for agent_id in agent_user_ids]
        }

        return self.requester.put(endpoint, data=data) 
    
    
    def group_call_center_bounced_calls(self, call_center_user_id: str, 
                                        updates: dict):
        """Update the bounced call settings of a single Call Center (CC)

        Args:
            call_center_user_id (str): Service user ID of the target CC.   
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/bounced-calls"
        
        updates["serviceUserId"] = call_center_user_id
        
        return self.requester.put(endpoint, data=updates) 
    
    
    def group_call_center_dnis_instance(self, call_center_user_id: str, 
                                        updates: dict):
        """Update a DNIS instance of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/dnis/instances"
        
        updates["serviceUserID"] = call_center_user_id

        return self.requester.put(endpoint, data=updates) 
    
    
    def group_call_center_forced_forwarding(self, call_center_user_id: str, 
                                            updates: dict):
        """Update the forced forwarding settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/forced-forwarding"
        
        updates["serviceUserID"] = call_center_user_id

        return self.requester.put(endpoint, data=updates) 
    

    def group_call_center_overflow(self, call_center_user_id: str, 
                                   updates: dict):
        """Update the overflow settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/overflow"
        
        updates["serviceUserID"] = call_center_user_id

        return self.requester.put(endpoint, data=updates) 
    

    def group_call_center_stranded_calls(self, call_center_user_id: str, 
                                         updates: dict):
        """Update the stranded calls settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/stranded-calls"
        
        updates["serviceUserID"] = call_center_user_id

        return self.requester.put(endpoint, data=updates) 
    

    def group_call_center_stranded_calls_unavailable(self, call_center_user_id: str, 
                                                     updates: dict):
        """Update the stranded calls unavailable settings of a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            updates (dict): Updates to apply to target CC.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/call-centers/stranded-calls-unavailable"
        
        updates["serviceUserID"] = call_center_user_id

        return self.requester.put(endpoint, data=updates) 
    
    
    def user_call_center_supervised_agents(self, call_center_user_id: str, 
                                           supervisor_user_id: str, agent_ids: list):
        """Update the list of agents a supervisor has in a single Call Center (CC).

        Args:
            call_center_user_id (str): Service user ID of the target CC. 
            supervisor_user_id (str): User ID of the supervisor.
            agent_ids (list): List of user IDs of agents to apply to supervisor.

        Returns:
            Dict: Superivor ID and list of agents they supervise.
        """
        
        endpoint = "/groups/call-centers/supervisors"
        
        data = {
            "serviceUserId": call_center_user_id,
            "supervisorUserId": supervisor_user_id,
            "supervisors": [{"userId": agent_id} for agent_id in agent_ids]
        }

        return self.requester.put(endpoint, data=data) 
    
    
    def user_call_center(self, user_id: str, updates: dict):
        """Update an agent's status in a Call Center (CC).

        Args:
            user_id (str): User ID of the target user.
            updates (dict): Updates to be applied to the user.

        Returns:
            Dict: Agents ACD status and status in each CC they are assigned to.
        """
        
        endpoint = "/users/call-center"
        
        updates["userId"] = user_id
           
        return self.requester.put(endpoint, data=updates)


    def user_call_center_agents_update(self, user_id: str, 
                                       call_center_service_ids: list):
        """Update the Call Centers (CC) a user is assigned to. 

        Args:
            user_id (str): User ID of the target user.
            call_center_service_ids (list): List of CC service user IDs to update the user with.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/user/call-centers/agents"
        
        data = {
            "agentUserId": user_id,
            "callCenters": [{"userId": call_center_id} for call_center_id in call_center_service_ids]
        }

        return self.requester.put(endpoint, data=data)   


    def user_call_center_agent_sign_out(self, user_id: str):
        """Sign the user out of their assigned Call Centers (CC).

        Args:
            user_id (str): User ID of the target user.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/user/call-centers/agents/sign-out"
        
        data = {
            "agentUserId": user_id,
        }

        return self.requester.put(endpoint, data=data)  
          
#CALL CONTROL
#CALL FORWARDING ALWAYS
#CALL FORWARDING ALWAYS SECONDARY
#CALL FORWARDING BUSY 
#CALL FORWARDING NO ANSWER
#CALL FORWARDING NOT REACHABLE
#CALL FORWARDING SELECTIVE
#CALL FORWARDING SETTINGS
#CALL NOTIFY
#CALL PARK
#CALL PICKUP
#CALL POLICIES
#CALL PROCESSING POLICIES

    def user_call_processing_policy(self, user_id: str, updates: dict):
        """
        Update the Call Processing Policies for a specified user. 

        Args:
            user_id (str): The user ID of the user whose call processing policies need updating.
            updates (dict): Updates to apply to the specified user. 
        
        Returns:
            Dict: Returns the updated call processing policies.
        """
        
        endpoint = "/users/call-processing-policy"

        updates["userId"] = user_id

        return self.requester.put(endpoint, data=updates)

#CALL RECORDING
#CALL RECORDS
#CALL TRANSFER
#CALL WAITING
#CALLING LINE ID BLOCKING OVERRIDE
#CALLING LINE ID DELIVERY BLOCKING
#CALLING NAME DELIVERY
#CALLING NAME RETRIEVAL
#CALLING NUMBER DELIVERY
#CALLING PARTY CATEGORY
#CALLING PLANS
#CALLBACKS
#CHARGENUMBER
#CLASSMARK
#CLONE
#COLLABORATE
#COMM PILOT CALL MANAGER
#COMM PILOT EXPRESS
#COMMON PHONE LIST
#COMMUNICATION BARRING
#COMMUNICATION BARRING USER
#CONNECTED LINE IDENTIFICATION
#CUSTOM CONTACT DIRECTORY
#DEPARTMENTS
#DEVICE POLICIES
#DEVICES


    def group_devices(self, service_provider_id: str, group_id: str, 
                      device_name: str, updates: dict):
        """Update a single device in a group. 

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = f"/groups/devices"
        
        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)  
    
    
    def service_provider_device(self, service_provider_id: str, device_name: str, 
                                updates: dict):
        """Update a single device in a Service Provider or Enterprise.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where device is located.
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/service-providers/devices"
        
        updates["serviceProviderId"] = service_provider_id
        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)
        
    
    def system_devices(self, device_name: str, updates: dict):
        """Update a single device in the Broadworks system.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
    
        endpoint = "/service-providers/devices"
        
        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)
        
    
    def system_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at the system level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/system/devices/files"
        
        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)
    
    
    def service_provider_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at Service Provider or
        Enterprise level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/service_provider/devices/files"

        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)
    

    def group_device_file(self, device_name: str, updates: dict):
        """Update a config file for a single device at Group level.

        Args:
            device_name (str): Device name of the target device.
            updates (dict): Updates to apply to the target device.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/devices/files"
        
        updates["deviceName"] = device_name
        
        return self.requester.put(endpoint, data=updates)
    
    
    def group_device_tags_profile(self, service_provider_id: str, group_id: str, 
                                  device_name: str, tags: list):
        """Update tags assigned to single device at group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            tags (dict): List of dictionaries tag name and value entries.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/groups/devices/profile"
        
        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceName": device_name,
            "tags": [
                {
                    "elements": tags
                }
            ]
        }
        
        return self.requester.put(endpoint, data=data)
    
    
    def group_device_tag(self, service_provider_id: str, group_id: str, device_name: str, 
                         tag_name: str, tag_value: str):
        """Update a single tag assigned to a device at group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/groups/devices/tags"
        
        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceName": device_name
        }     
        
        return self.requester.put(endpoint, data=data)
    
    
    def service_provider_device_tag(self, service_provider_id: str, device_name: str, 
                                    tag_name: str, tag_value: str):
        """Update a single tag assigned to a device at the Service Provider or Enterprise level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/service-providers/devices/tags"
        
        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "deviceName": device_name
        }     
        
        return self.requester.put(endpoint, data=data)
        
        
    def system_device_tag(self, device_name: str, tag_name: str, tag_value: str):
        """Update a single tag assigned to a device at the System level.

        Args:
            device_name (str): Device name of the target device.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/system/devices/tags"
        
        data = {
            "tagName": f"%{tag_name}%",
            "tagValue": tag_value,
            "deviceName": device_name
        }     
        
        return self.requester.put(endpoint, data=data)
    
    
    def group_device_type_file(self, service_provider_id: str, group_id: str, 
                               device_type: str, updates: dict):
        """Set config file for all devices of a specific type at the group level.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_type (str): The device type you would like to apply the changes to.
            updates (dict): Cupdates (dict): Updates to apply to the target device.
            
        Returns:
            None: This method does not return any specific value.
        """
    
        endpoint = "/groups/device-types/files"
        
        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["deviceType"] = device_type
        
        return self.requester.put(endpoint, data=updates)
        
    
    def group_device_type_tag(self, service_provider_id: str, group_id: str, 
                              device_type: str, tag_name: str, tag_value: str):
        """Update tags applied to device types at the Group level. 
                
        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            group_id (str): Group ID where target device is located.
            device_type (str): The target device type to apply the updates.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/groups/system/device-types/tags"
        
        data = {
            "tagName": tag_name,
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "deviceType": device_type
        }     
        
        return self.requester.put(endpoint, data=data)
    
    
    def service_provider_device_type_tag(self, service_provider_id: str, device_type: str, 
                                         tag_name: str, tag_value: str):
        """Update tags applied to device types at the Service Provider or Enterprise level. 
                
        Args:
            service_provider_id (str): Service Provider or Enterprise ID where Group is located.
            device_type (str): The target device type to apply the updates.
            tag_name (str): Name of the tag to add or update.
            tag_value (str): Value of tag to add or update.

        Returns:
            Dict: Python dictionary of the new state after updates have been applied.
        """
        
        endpoint = "/service-providers/device-types/tags"
        
        data = {
            "tagName": tag_name,
            "tagValue": tag_value,
            "serviceProviderId": service_provider_id,
            "deviceType": device_type
        }     
        
        return self.requester.put(endpoint, data=data)

#DIAL PLAN POLICY
#DIRECTED CALL PICKUP WITH BARGE IN
#DIRECTROUTE
#DN

    def group_dns_activate(self, service_provider_id: str, group_id: str, activated: bool, 
                           numbers: list):
        """Update activation state of a list of numbers assigned to a group.

        Args:
            service_provider_id (str): Service provider ID where the group is located. 
            group_id (str): Group ID where numbers are hosted.
            activated (bool): True to activate number and False to deactivate.
            numbers (list): List of target numbers to update. These must be strings and follow correct format.

        Returns:
            JSON: All numbers assigned to group with activation state.
        """
        
        endpoint = "/groups/dns"
        
        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [{"activated": activated, "min": num, "max": None} for num in numbers]
        }
        
        return self.requester.put(endpoint, data=data)
        

#DO NOT DISTURB

    def user_do_not_disturb(self, user_id: str, dnd_active: bool = False, ring_splash_active: bool = False ):
        """Updates a user's DND and Ring Splash status.

        Args:
            user_id (str): Target user id of user you would like to update the state of. 
            dnd_active (bool): True to enable DND and False to disable DND. Defaults to False.
            ring_splash_active (bool): True to enable Ring Splash and False to disable Ring Splash. Defaults to False.

        Returns:
            Dict: New DND and Ringsplash state of the user.
        """
        
        
        data = {
            "isActive":dnd_active,
            "ringSplash":ring_splash_active,
            "userId": user_id
        }
        
        endpoint = f"/users/do-not-disturb"
        
        return self.requester.put(endpoint, data)
        
        

#DOMAINS
#EMERGENCY NOTIFICATIONS
#EMERGENCY ZONES

    def group_emergency_zones(self, service_provider_id: str, group_id: str, is_active: bool=True, zone_rules: str=None, emergency_notification_email: str=None, ip_addresses: list=None):
        """Updates the Emergency Zone configuration in the group. 
       
        Args:
            service_provider_id (str): Service provider ID where the Emergency Zone to be updated exists.
            group_id (str): Group ID where the Emergency Zone to be updated exists.
            is_active (bool, optional): Whether the Emergency Zone service is active or not. Defaults to True
            zone_rules (str, optional): The rules of the Emergency Zone. This will either be "Prohibit all registrations and call originations" or "Prohibit emergency call originations".
            emergency_notification_email (str, optional): The email address where emergency call notifications should be sent. 
            ip_addresses (list, optional): A list of IP address ranges (dicts) to be added to the Emergency Zone. If the IP address to be applied is not a range, the min and max values should be the same.
            
        Returns:
            Dict: Updated Emergency Zone configuration.
        """

        endpoint = "/groups/emergency-zones"

        data = {
            "groupId": group_id, 
            "serviceProviderId": service_provider_id
        }

        if is_active:
            data["isActive"] = is_active
        if zone_rules:
            data["emergencyZonesProhibition"] = zone_rules
        if emergency_notification_email:
            data["sendEmergencyCallNotifyEmail"] = True
            data["emergencyCallNotifyEmailAddress"] = emergency_notification_email
        if ip_addresses:
            data["ipAddresses"] = ip_addresses
        
        return self.requester.put(endpoint, data)        


#ENTERPRISE TRUNKS
#EXECUTIVE
#EXECUTIVE ASSISTANT
#EXTENSIONS
#EXTERNAL CALLING LINE ID DELIVERY
#EXTERNAL CUSTOM RINGBACK
#FAX MESSAGING
#FEATURE ACCESS CODES
#FLEXIBLE SEATING
#GROUP PAGING
#GROUPS
#GROUP NAVIGATION
#HOTELING GUEST
#HOTELING HOST
#HUNT GROUPS
    
    def group_hunt_groups_status(self, service_user_ids: list, status: bool =True):
        """Updates a list of Hunt Groups (HG) status to either active or inactive.

        Args:
            service_user_ids (list): List of service user IDs of target HGs.
            status (bool, optional): Status to apply to target HGs. Defaults to True.

        Returns:
            None: This method does not return any specific value.
        """
    
        endpoint = f"/groups/hunt-groups/status"
        
        data = {     
            "instances": [{'serviceUserId': service_user_id, 'isActive': status} 
                          for service_user_id in service_user_ids]
        }
        
        return self.requester.put(endpoint, data=data)
        
    
    def group_hunt_group(self, service_provider_id: str, group_id: str, service_user_id: str, updates: dict):
        """Update a Hunt Group's (HG) settings.

        Args:
            service_provider_id (str): Service provider ID of where the group that hosts the HG is located.
            group_id (str): Group ID of where the HG is located.
            service_user_id (str): Target service user ID of the HG.
            updates (dict): Updates to be applied to HG.

        Returns:
            None: This method does not return any specific value.
        """
    
        endpoint = f"/groups/hunt-groups"
        
        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["serviceUserId"] = service_user_id
        if not updates.get("serviceInstanceProfile"):
            updates["serviceInstanceProfile"] = {}          
        
        return self.requester.put(endpoint, data=updates)
            
        
    def group_hunt_group_weighted_call_distribution(self, service_provider_id: str, group_id, service_user_id: str, 
                                                    agents: list):
        """Update the Weighted Call Distribution (WCD) between users in a Hunt Group (HG).

        Args:
            service_provider_id (str): Service provider ID where the group is located. 
            group_id (_type_): Group ID where the HG is located.
            service_user_id (str): Service user ID of the target HG.
            agents (list): Updates of WCD to be applied to HG.

        Raises:
            AOInvalidWeighting: The WCD must equal 100 if it does not this error will be returned.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = f"/groups/hunt-groups/weighted-call-distribution"
        
        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "serviceUserId": service_user_id,
            "agents": agents
        }
        
        max_weights = 100
        assigned_weight = 0
        
        for agent in agents:
            assigned_weight += agent["weight"]                
        if not assigned_weight == max_weights:
            raise OSInvalidWeighting
      
        return self.requester.put(endpoint, data=data)
        
        
        
        

#IN CALL SERVICE ACTIVATION
#INSTANT GROUP CALL
#INTEGRATED IMP
#INTERCEPT
#INTERNAL CALLING LINE ID DELIVERY
#LANGUAGES
#LEGACY AUTOMATIC CALLBACK
#MALICIOUS CALL TRACE
#MEDIA
#MEET-ME CONFERENCING
#MUSIC ON HOLD
#MWI DELIVERY TO MOBILE ENDPOINT
#NETWORK CLASS OF SERVICE
#NIGHT FORWARDING
#NUMBER PORTABILITY ANNOUNCEMENT
#NUMBER PORTABILITY QUERY
#NUMBERS
#OUTLOOK INTEGRATION
#PASSCODE RULES
#PASSWORD GENERATE
#PASSWORD RULES
#PERSONAL PHONE LIST
#PHONE DIRECTORY
#PHYSICAL LOCATION
#POLYCOM PHONE SERVICES
#PRE ALERTING
#PREFERRED CARRIER
#PREPAID
#PRIMARY ENDPOINT ADVANCED SETTING
#PRIORITY ALERT
#PRIVACY
#PUSH REGISTRATION
#PUSH TO TALK
#REGISTRATION
#REMOTE OFFICE
#REPORTS
#RESELLERS
#ROUTE LIST
#ROUTING PROFILE
#SCHEDULES
#SECURITY CLASSIFICATION
#SELECTIVE CALL ACCEPTANCE
#SELECTIVE CALL REJECTION
#SEQUENTIAL RING
#SERIES COMPLETION
#SERVICE PACKS
#SERVICE PROVIDERS
#SERVICES

    def user_services(self, user_id: str, services: list =None, service_packs: list =None, assigned: bool =True):
        """Update the services assigned to a user. NOT service/feature packs.

        Args:
            user_id (str): User ID of the target user.
            services (list): List of services to be applied to user.
            service_packs (list): List of service packs to be applied to user.
            assigned (bool, optional): Assign (True) or unassign(False). Defaults to True.

        Returns:
            Dict: User services assigned to the user. 
        """
        
        endpoint = f"/users/services"
        
        data = {
            "userId": user_id
        }
        
        if services:
            data["userServices"] = [{'serviceName': service, 'assigned': assigned} for service in services]
        if service_packs:
            data["servicePackServices"] = [{'serviceName': service_pack, 'assigned': assigned} for service_pack in service_packs]
        return self.requester.put(endpoint, data=data)
    
    def user_service_settings(self, user_id: str, settings: dict):
        """Updates specific service settings for a given user.
        This function allows you to modify one or more service settings associated with a particular user.

        Args:
            user_id (str): The ID of the target user
            settings (dict): A dictionary containing the new settings to be applied. The structure of this dictionary should mirror the API's expected format for updating service settings.

        Returns:
            Dict: A dictionary representing the updated service settings for the specified user.
        """

        endpoint = f"/users/services/settings"

        data = {
            "userId": user_id,
            **settings
        }

        return self.requester.put(endpoint, data=data)

#SHARED CALL APPEARANCE
#SILENT ALERTING
#SIMULTANEOUS RING PERSONAL
#SMARTY ADDRESS
#SPEED DIAL 100
#SPEED DIAL 8
#STATES AND PROVINCES
#SYSTEM
#TERMINATING ALTERNATE TRUNK IDENTITY
#THIRD PARTY VOICE MAIL SUPPORT
#THIRD PARTY EMERGENCY CALLING
#TIME ZONES
#TRUNK GROUPS

    def group_trunk_groups_call_capacity(self, service_provider_id: str, group_id: str, max_active_calls: int=None,
                                         bursting_max_active_calls: int=None, number_of_bursting_btlus: int=None):
        """
        Updates the trunking call capacity in the specified group. 
        NOTE: The max available active calls cannot be changed here. Please see service_providers_trunk_group_call_capacity to update this.

        Args:
            service_provider_id (str): Service provider ID where the target group is built
            group_id (str): Group ID whose trunk group call capacity needs updating
            max_active_calls (int, optional): The max active calls for the group. 
            bursting_max_active_calls (int, optional): The bursting max active calls for the group.
            number_of_bursting_btlus (int, optional): The number of Business Trunking License Units for bursting. 

        Returns: 
            Dict: Returns the updated state of the trunk group call capacity.
        """

        endpoint = "/groups/trunk-groups/call-capacity"
        
        updates = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id
        }

        if max_active_calls:
            updates["maxActiveCalls"] = max_active_calls
        if bursting_max_active_calls:
            updates["burstingMaxActiveCalls"] = bursting_max_active_calls
        if number_of_bursting_btlus:
            updates["numberOfBurstingBTLUs"] = number_of_bursting_btlus

        return self.requester.put(endpoint, data=updates)


    def group_trunk_group(self, service_provider_id: str, group_id: str, trunk_group_name: str, updates: dict):
        """
        Updates trunk group (TG) information.

        Args: 
            service_provider_id (str): Service provider ID where the target group is built
            group_id (str): Group ID whose trunk group call capacity needs updating
            trunk_group_name (str): The name of the trunk group that is being updated. 
            updates (dict): Updates to be applied to the TG. 

        Returns:
            Dict: Returns the updated Trunk Group profile.
        """

        endpoint = "/groups/trunk-groups"

        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["name"] = trunk_group_name

        return self.requester.put(endpoint, data=updates)


    def service_providers_trunk_group_call_capacity(self, service_provider_id: str, max_active_calls: int, bursting_max_active_calls: int =None):
        """
        Updates the max active calls and the bursting max active calls for the given service provider.

        Args: 
            service_provider_id (str): Service provider ID for which the max active calls needs to be updated 
            updates (dict): The updates to be applied to the service provider's trunking call capacity
        
        Returns:
            Dict: Returns the updated call capacity for the service provider. 
        """

        endpoint = "/service-providers/trunk-groups/call-capacity"
        
        data = {
            "maxActiveCalls": max_active_calls,
            "serviceProviderId": service_provider_id
        }
        
        if bursting_max_active_calls:
            data["burstingMaxActiveCalls"] = bursting_max_active_calls

        return self.requester.put(endpoint, data=data)

#TWO STAGE DIALING
#USERS

    def users_bulk(self, users: list, updates: dict):
        """
        Updates specified list of User's options, such as extension, name and etc.

        Note: Available options to change can be seen through: get.user_by_id()

        Args: 
            users (list): List of specified User IDs to update
            updates (dict): The updates to be applied to the list of Users e.g {"extension":"9999"}
        
        Returns:
            Dict: Returns the changes made including the list of User IDs and updates.
        """
        endpoint = "/users/bulk"
        
        target_users = [{"userId": user} for user in users]
        
        data = {
            "users": target_users,
            "data": updates 
        }
        
        return self.requester.put(endpoint, data=data)


    def user(self, service_provider_id: str, group_id: str, user_id: str, updates: dict):
        """
        Updates specified User's options, such as extension, name and etc.

        Note: Available options to change can be seen through: get.user_by_id()

        Args: 
            service_provider_id (str): 
            updates (dict): The updates to be applied Target Service Provider where group is located
            group_id (str): Target Group ID where user is located
            user_id (str): Target User IDto the list of Users e.g {"extension":"9999"}
        
        Returns:
            Dict: Returns the changes made including User ID and updates.
        """
        endpoint = "/users"
        
        updates["serviceProviderId"] = service_provider_id
        updates["groupId"] = group_id
        updates["userId"] = user_id
        
        return self.requester.put(endpoint, data=updates)
        
    
    def user_portal_passcode(self, user_id: str, new_passcode: str):
        """Updates the specified User's portal passcode.

        Args:
            user_id (str): User ID of the target user you would like to change the portal passcode for. 
            new_passcode (int): New portal passcode to set for the target user.

        Raises:
            AOInvalidCode: If code is less than 4 or higher than 6.

        Returns:
            None: This method does not return any specific value.
        """
        
        if len(new_passcode) < 4 or len(new_passcode) > 6:
            raise OSInvalidCode
        
        endpoint = "/users/portal-passcode"
        
        data = {
            "userId": user_id,
            "newPasscode": new_passcode
        }
        
        return self.requester.put(endpoint, data=data)
   
    
#USER CUSTOM RINGBACK
#VIDEO ADD ON
#VIRTUAL ON-NET ENTERPRISE EXTENSIONS
#VOICE MESSAGING
#VOICE PORTAL CALLING
#ZONE CALLING RESTRICTIONS
#ODIN BRANDING
#ODIN CALLBACKS
#ODIN CONNECTORS
#ODIN EMAIL
#ODIN EVENTS
#ODIN INVENTORY
#ODIN REPORTS
#ODIN SETTINGS
#ODIN STATUS
#ODIN SSO
#ODIN SSO ALTERNATE USER IDS
#ODIN TASKS
#ODIN UI
#ODIN VIEWABLE PACKS
#ODIN WEBHOOKS
#ODIN AUDIT
#ODIN IMPORTS
#ODIN EXPORTS
#XSI
#SIP AUTHENTICATION
#POLICY
#PASSWORD RESET
#PARTNERS
#USER UTILITIES
#ODIN TASKS COPY
#ODIN CONNECTORS
#LOCALES



from ...store import broadwork_entities as bre
from .report_entities import external_number

# nodes needed for graph generation
NODES = [] 
MAX_LEVEL = 30

def call_flow_module(node: object, data_store: object):
    
    # create mapping for easier searchability when gathering nodes
    data_store.build_number_mapping()
    
    # traverses the enities that forward to eachother and adds all to NODES
    _traverse_connecting_entities(node, data_store)
    
    return NODES
    

def _traverse_connecting_entities(entity: object, data_store: object, visited= None):
    if visited is None:
        visited = []
        
    if entity in visited:
        return
    else:
        visited.append(entity) 
    
    try:
        if isinstance(entity, bre.User):
            # Checks to see if the number is external first and creates external number object for parsing later
            if data_store.number_mapping.get(entity.call_forwarding_always) == None and len(str(entity.call_forwarding_always)) >= 3 and str(entity.call_forwarding_always)[2].isdigit():
                entity.call_forwarding_always = external_number(str(entity.call_forwarding_always))
            else:
                entity.call_forwarding_always = data_store.number_mapping.get(entity.call_forwarding_always)
                if entity.call_forwarding_always and entity.call_forwarding_always not in visited: 
                    _traverse_connecting_entities(entity.call_forwarding_always, data_store, visited)
            
            if data_store.number_mapping.get(entity.call_forwarding_busy) == None and len(str(entity.call_forwarding_busy)) >= 3 and str(entity.call_forwarding_busy)[2].isdigit():
                entity.call_forwarding_busy = external_number(str(entity.call_forwarding_busy))
            else:
                entity.call_forwarding_busy = data_store.number_mapping.get(entity.call_forwarding_busy)
                if entity.call_forwarding_busy and entity.call_forwarding_busy not in visited:
                    _traverse_connecting_entities(entity.call_forwarding_busy, data_store, visited)
            
            if data_store.number_mapping.get(entity.call_forwarding_no_answer) == None and len(str(entity.call_forwarding_no_answer)) >= 3 and str(entity.call_forwarding_no_answer)[2].isdigit():
                entity.call_forwarding_no_answer = external_number(str(entity.call_forwarding_no_answer))
            else:
                entity.call_forwarding_no_answer = data_store.number_mapping.get(entity.call_forwarding_no_answer)
                if entity.call_forwarding_no_answer and entity.call_forwarding_no_answer not in visited:
                    _traverse_connecting_entities(entity.call_forwarding_no_answer, data_store, visited)
            
            if data_store.number_mapping.get(entity.call_forwarding_not_reachable) == None and len(str(entity.call_forwarding_not_reachable)) >= 3 and str(entity.call_forwarding_not_reachable)[2].isdigit():
                entity.call_forwarding_not_reachable = external_number(str(entity.call_forwarding_not_reachable))
            else:
                entity.call_forwarding_not_reachable = data_store.number_mapping.get(entity.call_forwarding_not_reachable)
                if entity.call_forwarding_not_reachable and entity.call_forwarding_not_reachable not in visited:
                    _traverse_connecting_entities(entity.call_forwarding_not_reachable, data_store, visited)
        
        if isinstance(entity, bre.CallCenter):
            if entity.bounced_calls_enabled:
                # Checks to see if the number is external first and creates external number object for parsing later 
                if data_store.number_mapping.get(str(entity.bounced_calls_transfer_to_phone_number)) == None and len(str(entity.bounced_calls_transfer_to_phone_number)) >= 3 and str(entity.bounced_calls_transfer_to_phone_number)[2].isdigit():
                    entity.bounced_calls_transfer_to_phone_number = external_number(str(entity.bounced_calls_transfer_to_phone_number))
                else:
                    entity.bounced_calls_transfer_to_phone_number = data_store.number_mapping.get(str(entity.bounced_calls_transfer_to_phone_number))
                    if entity.bounced_calls_transfer_to_phone_number and entity.bounced_calls_transfer_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.bounced_calls_transfer_to_phone_number, data_store, visited)
            
            if entity.overflow_calls_action:
                if data_store.number_mapping.get(str(entity.overflow_calls_transfer_to_phone_number)) == None and len(str(entity.overflow_calls_transfer_to_phone_number)) >= 3 and str(entity.overflow_calls_transfer_to_phone_number)[2].isdigit():
                    entity.overflow_calls_transfer_to_phone_number = external_number(str(entity.overflow_calls_transfer_to_phone_number))
                else:
                    entity.overflow_calls_transfer_to_phone_number = data_store.number_mapping.get(str(entity.overflow_calls_transfer_to_phone_number))
                    if entity.overflow_calls_transfer_to_phone_number and entity.overflow_calls_transfer_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.overflow_calls_transfer_to_phone_number, data_store, visited)

            if entity.stranded_calls_action:
                if data_store.number_mapping.get(str(entity.stranded_calls_transfer_to_phone_number)) == None and len(str(entity.stranded_calls_transfer_to_phone_number)) >= 3 and str(entity.stranded_calls_transfer_to_phone_number)[2].isdigit():
                    entity.stranded_calls_transfer_to_phone_number = external_number(str(entity.stranded_calls_transfer_to_phone_number))
                else:
                    entity.stranded_calls_transfer_to_phone_number = data_store.number_mapping.get(str(entity.stranded_calls_transfer_to_phone_number))
                    if entity.stranded_calls_transfer_to_phone_number and entity.stranded_calls_transfer_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.stranded_calls_transfer_to_phone_number, data_store, visited)
        
            if entity.stranded_call_unavailable_action:
                if data_store.number_mapping.get(str(entity.stranded_call_unavailable_transfer_to_phone_number)) == None and len(str(entity.stranded_call_unavailable_transfer_to_phone_number)) >= 3 and str(entity.stranded_call_unavailable_transfer_to_phone_number)[2].isdigit():
                    entity.stranded_call_unavailable_transfer_to_phone_number = external_number(str(entity.stranded_call_unavailable_transfer_to_phone_number))
                else:
                    entity.stranded_call_unavailable_transfer_to_phone_number = data_store.number_mapping.get(str(entity.stranded_call_unavailable_transfer_to_phone_number))
                    if entity.stranded_call_unavailable_transfer_to_phone_number and entity.stranded_call_unavailable_transfer_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.stranded_call_unavailable_transfer_to_phone_number, data_store, visited)
        
        if isinstance(entity, bre.HuntGroup):
            if entity.forward_after_timeout_enabled:
                if data_store.number_mapping.get(entity.no_answer_forward_to_phone_number) is None and len(str(entity.no_answer_forward_to_phone_number)) >= 4 and str(entity.no_answer_forward_to_phone_number)[2].isdigit():
                    entity.no_answer_forward_to_phone_number = external_number(str(entity.no_answer_forward_to_phone_number))
                else:
                    entity.no_answer_forward_to_phone_number = data_store.number_mapping.get(str(entity.no_answer_forward_to_phone_number))
                    if entity.no_answer_forward_to_phone_number and entity.no_answer_forward_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.no_answer_forward_to_phone_number, data_store, visited)
            
            if entity.call_forward_not_reachable_enabled:
                if data_store.number_mapping.get(str(entity.call_forward_not_reachable_transfer_to_phone_number)) == None and len(str(entity.call_forward_not_reachable_transfer_to_phone_number)) >= 3 and str(entity.call_forward_not_reachable_transfer_to_phone_number)[2].isdigit():
                    entity.call_forward_not_reachable_transfer_to_phone_number = external_number(str(entity.call_forward_not_reachable_transfer_to_phone_number))
                else:
                    entity.call_forward_not_reachable_transfer_to_phone_number = data_store.number_mapping.get(str(entity.call_forward_not_reachable_transfer_to_phone_number))
                    if entity.call_forward_not_reachable_transfer_to_phone_number and entity.call_forward_not_reachable_transfer_to_phone_number not in visited:
                        _traverse_connecting_entities(entity.call_forward_not_reachable_transfer_to_phone_number, data_store, visited)     
        
        if isinstance(entity, bre.AutoAttendant):
            for key in entity.business_hours_menu.keys:
                if "Transfer" in key.action:
                    if data_store.number_mapping.get(str(key.phone_number)) == None and len(str(key.phone_number)) >= 3 and str(key.phone_number)[2].isdigit():
                        key.phone_number = external_number(str(key.phone_number))
                    else:
                        key.phone_number = data_store.number_mapping.get(str(key.phone_number))
                        if key.phone_number and key.phone_number not in visited:
                            _traverse_connecting_entities(key.phone_number, data_store, visited)
            
            for key in entity.after_hours_menu.keys:
                if "Transfer" in key.action:
                    if data_store.number_mapping.get(str(key.phone_number)) == None and len(str(key.phone_number)) >= 3 and str(key.phone_number)[2].isdigit():
                        key.phone_number = external_number(str(key.phone_number))
                    else:
                        key.phone_number = data_store.number_mapping.get(str(key.phone_number))
                        if key.phone_number and key.phone_number not in visited:
                            _traverse_connecting_entities(key.phone_number, data_store, visited)
    except TypeError:
        # this object is further up the tree and a loop has occured - Move one and come back 
        pass

    NODES.append(entity)  # Append outside of the if condition




def generate_timestamp():
    """Generates a timestamp of the current time in string format.

    Returns:
        str: Timestamp of existing time when called.
    """
    from datetime import datetime
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


import json
from tqdm import tqdm


def main(api, service_provider_id: str, group_id: str):
    """ Audits a group for chargeable services this will return all feature packs 
    assigned and count for all broadwork entities such as users, call centers, hunt groups etc.
    Additionaly this will return all DID/DDI's and their active status.

    :param service_provider_id: Service provider or Enterprise ID that contains group being audited.
    :param group_id: Group that is to be audited.

    :return r:
    """

    # all Services
    service_report = api.get.group_services(group_id, service_provider_id)

    assigned_user_services = []
    assigned_group_services = []
    assigned_service_pack_services = []

    for us in tqdm(service_report["userServices"], desc="Analysing User Services..."):
        if us["usage"] > 0:
            del us["authorized"]
            del us["assigned"]
            del us["limited"]
            del us["quantity"]
            del us["licensed"]
            del us["allowed"]
            del us["userAssignable"]
            del us["groupServiceAssignable"]
            del us["tags"]
            del us["alias"]

            users = api.get.group_services_user_assigned(
                group_id, service_provider_id, us["serviceName"], "serviceName")
            userIDs = [u["userId"] for u in users["users"]]
            us["users"] = userIDs

            assigned_user_services.append(us)

    for gs in tqdm(service_report["groupServices"], desc="Analysing Group Services..."):
        if gs["usage"] > 0:
            del gs["authorized"]
            del gs["assigned"]
            del gs["limited"]
            del gs["quantity"]
            del gs["licensed"]
            del gs["allowed"]
            del gs["instanceCount"]
            del gs["alias"]
            assigned_group_services.append(gs)

    for sps in tqdm(service_report["servicePackServices"], desc="Analysing Service Packs..."):
        if sps["usage"] > 0:
            del sps["authorized"]
            del sps["assigned"]
            del sps["limited"]
            del sps["allowed"]
            del sps["serviceName"]
            del sps["quantity"]
            del sps["alias"]

            users = api.get.group_services_user_assigned(
                group_id, service_provider_id, sps["servicePackName"], "servicePackName")
            userIDs = [u["userId"] for u in users["users"]]
            sps["users"] = userIDs

            assigned_service_pack_services.append(sps)

    # Group DNs
    dn_report = api.get.group_dns(service_provider_id, group_id)
    all_dns = {
        "assigned": {
            "activated":[],
            "deactivated": []
        },
        "unassigned": {
            "activated": [],
            "deactivated": []
        }
    }

    for dn in tqdm(dn_report["dns"], desc="Analysing Group DNs..."):
        if dn["assigned"] and dn["activated"]:
            all_dns["assigned"]["activated"] += dn["list"]
        elif dn["assigned"] and not dn["activated"]:
            all_dns["assigned"]["deactivated"] += dn["list"]
        elif not dn["assigned"] and dn["activated"]:
            all_dns["unassigned"]["activated"] += dn["list"]
        elif not dn["assigned"] and not dn["activated"]:
            all_dns["unassigned"]["deactivated"] += dn["list"]

    total_assigned_activated = len(all_dns["assigned"]["activated"])
    total_assigned_deactivated = len(all_dns["assigned"]["deactivated"])
    total_unassigned_activated = len(all_dns["unassigned"]["activated"])
    total_unassigned_deactivated = len(all_dns["unassigned"]["deactivated"])
    total_dns = total_assigned_activated + total_assigned_deactivated + total_unassigned_activated + total_unassigned_deactivated

    all_dns["totalDNs"] = total_dns
    all_dns["assigned"]["totalAssignedDNs"] = total_assigned_activated + total_assigned_deactivated
    all_dns["unassigned"]["totalUassignedDNs"] = total_unassigned_activated + total_unassigned_deactivated

    # Group Detail
    group_detail = api.get.group(service_provider_id,group_id)

    #Trunking detail
    try:
        trunk_detail = api.get.group_trunk_groups_call_capacity(service_provider_id,group_id)
        del trunk_detail["serviceProviderId"]
        del trunk_detail["groupId"]
    except:
        trunk_detail = []

    group_audit = {
        "groupDetail": group_detail,
        "licenceBreakdown": {
            "userServices": assigned_user_services,
            "groupServices": assigned_group_services,
            "servicePackServices": assigned_service_pack_services
        },
        "groupDNs": all_dns,
        "groupTrunking": trunk_detail

    }
    
    return group_audit





def main(api, current_service_provider_id: str, current_group_id: str, target_service_provider_id: str, 
         target_group_id: str, start_of_range_number: str, end_of_range_number: str = None) -> bool:
    """ Moves singular or a range of numbers from one group to another located on the same Broadwork instance.
    """
    
    print(f"Removing numbers from {current_group_id}.")
    # delete number from group
    api.delete.group_dns(
        current_service_provider_id,
        current_group_id,
        start_of_range_number = start_of_range_number,
        end_of_range_number = end_of_range_number
    )
    
    # check if sp/ent are the same if not number needs to be removed to sys level
    if not current_service_provider_id == target_service_provider_id:
        
        print(f"Removing numbers from {current_service_provider_id}.")
        # remove from sp/ent
        api.delete.service_provider_dns(
            current_service_provider_id,
            start_of_range_number = start_of_range_number,
            end_of_range_number = end_of_range_number
        )
        
        print(f"Adding number to SP/ Ent: {target_service_provider_id} Group: {target_group_id}.")
        # assign to new group
        api.post.group_dns_assign_bulk(
            target_service_provider_id,
            target_group_id,
            start_of_range_number = start_of_range_number,
            end_of_range_number = end_of_range_number
        )
    else:
    # only when group is in the same sp/ent    
        
        print(f"Adding number to {target_group_id}.")
        # assign to new group
        api.post.group_dns(
            target_service_provider_id,
            target_group_id,
            start_of_range_number = start_of_range_number,
            end_of_range_number = end_of_range_number
        )
    
    return True


import json
from tqdm import tqdm

def remove_excess_details(aa_cc_hg: dict):
        del aa_cc_hg["name"]
        del aa_cc_hg["phoneNumber"]
        del aa_cc_hg["extension"]
        del aa_cc_hg["department"]
        del aa_cc_hg["isActive"]
        
        if aa_cc_hg["typeTag"] == "AA":
            del aa_cc_hg["video"]
        elif aa_cc_hg["typeTag"] == "CC":
            del aa_cc_hg["video"]
            del aa_cc_hg["policy"]
        elif aa_cc_hg["typeTag"] == "HG":
            del aa_cc_hg["serviceProviderId"]
            del aa_cc_hg["groupId"]
            del aa_cc_hg["policy"]
        
def main(api, service_provider_id: str, group_id: str):
    """
    This script returns the services assigned to Auto Attendants, 
    Call Centres, and Hunt Groups. Only services are applied to these 
    entities and there are scenarios one would need to focus services 
    assigned to these entities.
    

    :param service_provider_id: Service Provider ID or Enterprise ID containing the Group ID.
    :param group_id: Group ID to generate the report for.

    :return JSON: A JSON formatted report of service packs assigned to AA, CC, and HG. 
    """  

    print("aa_cc_hg_audit start.")
    return_data = {
        "autoAttendants": [],
        "callCenters": [],
        "huntGroups": []
    }
    auto_attendants = api.get.auto_attendants(service_provider_id, group_id)
    call_centers = api.get.group_call_centers(service_provider_id, group_id)
    hunt_groups = api.get.group_hunt_groups(service_provider_id, group_id)
    
    for aa in tqdm(auto_attendants, desc="Analysing Auto Attendants"):
        aa["typeTag"] = "AA"
        remove_excess_details(aa)

    for cc in tqdm(call_centers, desc="Analysing Call Centers"):
        cc["typeTag"] = "CC"
        remove_excess_details(cc)
        
    for hg in tqdm(hunt_groups, desc="Analysing Hunt Groups"):
        hg["typeTag"] = "HG"
        remove_excess_details(hg)
        
    aa_cc_hgs = auto_attendants + call_centers + hunt_groups
    
    for aa_cc_hg in tqdm(aa_cc_hgs, desc="Fetching User Services"):
        response = api.get.user_services_assigned(aa_cc_hg["serviceUserId"])
        aa_cc_hg["services"] = response["userServices"]
        
        if aa_cc_hg["typeTag"] == "AA":
            del aa_cc_hg["typeTag"]
            return_data["autoAttendants"].append(aa_cc_hg)
        elif aa_cc_hg["typeTag"] == "CC":
            del aa_cc_hg["typeTag"]
            return_data["callCenters"].append(aa_cc_hg)
        elif aa_cc_hg["typeTag"] == "HG":
            del aa_cc_hg["typeTag"]
            return_data["huntGroups"].append(aa_cc_hg)
        
    return return_data
        
    




def main(api, service_provider_id, group_id, user_id, device_type, 
         email, primary_device, webex_feature_pack_name, enable_integarated_imp):

	print("Start.\n")
	# Update the email & alt user id
	email_alt_userid = {
		"emailAddress": email,
    	"alternateUserId": [
        	{
            	"alternateUserId": email,
            	"description": "Webex"
        	}
    	]
    } 
 
	try:
		api.put.user(
			service_provider_id,
			group_id, 
			user_id,
			updates=email_alt_userid
		)
		print("Updated email and alt user ID.")
	except Exception as e:
		print(f"\tERROR: Failed to update users email and alt user ID. Detail: {e}")
	
	# Assign feature pack
	try:
		if webex_feature_pack_name:
			api.put.user_services(
				user_id=user_id, 
				service_packs=[webex_feature_pack_name]
			)
		print(f"Added feature: {webex_feature_pack_name}")
	except Exception as e:
		print(f"\tERROR: Failed to add feature - {webex_feature_pack_name}. Detail: {e}")

	# enable IMP in service settings
	if enable_integarated_imp:
		enable_IMP = {'Integrated IMP': {'isActive': True}}
		try:
			api.put.user_service_settings(user_id=user_id, settings=enable_IMP)
			print("Enabled integrated IMP.")
		except Exception as e:
			print(f"\tERROR: Failed to enable Integrated IMP. Detail: {e}")
  
	# build device
	device_name = f"{user_id.split('@')[0]}_WBX"
	device_password = api.get.password_generate(service_provider_id, group_id)["password"]

	device_payload = {
		"useCustomUserNamePassword": "true",
		"accessDeviceCredentials": {
			"userName": device_name,
			"password": device_password
		},
		"userName": device_name
	}

	try:
		api.post.group_device(
			service_provider_id=service_provider_id, 
			group_id=group_id, 
			device_name=device_name, 
			device_type=device_type, 
			payload=device_payload
		)
		print("Built device.")
	except Exception as e:
		print(f"\tERROR: Failed to build device. Detail: {e}")

	# Add device to user based on primary flag - JP
	if primary_device:
		primary_device_configuration = {
			"endpointType": "accessDeviceEndpoint",
			"accessDeviceEndpoint": {
				"accessDevice": {
				"deviceType": device_type,
				"deviceName": device_name,
				"serviceProviderId": service_provider_id,
				"groupId": group_id,
				"deviceLevel": "Group",
			},
			"linePort": user_id,
			}
		}
		try:
			api.put.user(service_provider_id, group_id, user_id, primary_device_configuration)
			print("Added device to user as primary.")
		except Exception as e:
			print(f"\tERROR: Failed to add device as primary. Detail: {e}")
	else:
		try:
			api.post.user_shared_call_appearance_endpoint(
				user_id,
				user_id.replace("@", "_WBX@"),
				device_name
			)
			print("Added device as shared call appearance.")
		except Exception as e:
			print(f"\tERROR: Failed to add devie as shared call appearance. Detail {e}")
  
	# Get webex password 
	try:
		password = api.get.password_generate(service_provider_id, group_id)["password"]
		api.put.user_web_authentication_password(user_id, password)
		print("Set webex password.")
	except Exception as e:
		print(f"\tERROR: Failed to set webex password. Detail {e}")

	# Return formatted data
	webex_user_details = {
	"username": email,
	"password": password,
	"primary_device": primary_device,
	"device_type": device_type
	}

	print("\nEnd.")
	return webex_user_details




def main(api, current_service_provider_id: str, current_group_id: str,
         start_of_range_number: str, end_of_range_number: str = None) -> bool:
    """Removes a singular or a range of numbers from a Broadworks instance.
    """
    
    print(f"Removing numbers from {current_group_id}.")
    # delete number from group
    api.delete.group_dns(
        current_service_provider_id,
        current_group_id,
        start_of_range_number = start_of_range_number,
        end_of_range_number = end_of_range_number
    )

    print(f"Removing numbers from {current_service_provider_id}.")
    # remove from sp/ent
    api.delete.service_provider_dns(
        current_service_provider_id,
        start_of_range_number = start_of_range_number,
        end_of_range_number = end_of_range_number
    )

    return True


""" Library exceptions.
"""

class OSError(Exception):
    """ Odin Api Exceptions
    """

    def __str__(self) -> str:
        return f"I dont think you can be trusted in a combat situation"


class OSApiAuthenticationFail(OSError):
    """
    Raised when api fails to authenticate
    """

    def __str__(self) -> str:
        return f"Failed to authenticate. Check username, password, and url." 


class OSRequestTypeError(OSError):
    """ Raised when unsupport request type is given. 
    """

    def __str__(self) -> str:
        return f"Non-supported request type, supported: GET, POST, PUT, DELETE."
    

class OSObjectParseError(OSError):
    """ Raised when parsing Broadworks Entity fails. 
    """

    def __str__(self) -> str:
        return f"Parsing Broadwork Entity failed."

    
class OSUnsupportedFilter(OSError):
    """ Raised when user requests to filter on unsupported filter
    """

    def __str__(self) -> str:
        return f"Unsupported filter. Supported: macAddress, lastName," \
            f"firstName, dn, emailAddress, userId, extension"

    
class OSAliasNotFound(OSError):
    """ Raised when alias is not found in Broadowks Group. 
    """

    def __str__(self) -> str:
        return f"Alias not found, it either does not exist or check alias."

    
class OSSessionRefreshFail(OSError):
    """ Raised when refreshing session fails.
    """

    def __str__(self) -> str:
        return f"Refreshing sesion failed. Check credentials are valid and " \
            f"token has not yet expired. If expired request another."

            
class OSLogoutFailed(OSError):
    """ Raised when logout attempt failed.
    """

    def __str__(self) -> str:
        return f"Failed to logout, session still valid. Please try again."

    
class OSFailedToLocateSession(OSError):
    """ Raised when user attempts to get session details but session cant be found.
    """

    def __str__(self) -> str:
        return f"Session details not found. Check token is valid and not exppired."

    
class OSInvalidCode(OSError):
    """ Raised when code is less than 4 and higher than 6.
    """

    def __str__(self) -> str:
        return f"Code needs to be between 4 and 6 digits."

    
class OSInvalidWeighting(OSError):
    """ Raised when invalid weighted call distribution set.
    """

    def __str__(self) -> str:
        return f"Weighted call distribution is not equal to 100. This weight must add up to 100."
    

class OSInvalidData(OSError):
    """ Raised when data used in request is invalid or incomplete.
    """

    def __str__(self) -> str:
        return f"Data invalid or incomplete, please check data passed to method is correct."


class OSInvalidBroadworkService(OSError):
    """ Raised when service given by user is not a valid license of Broadworks.
    """

    def __str__(self) -> str:
        return f"Service invalid, please check the services you have given are valid Broadwork services."
    
 
class OSInvalidPasswordType(OSError):
    """ Raised when password requested is invalid or not supported.  
    """

    def __str__(self) -> str:
        return f"Invalid or unsupported password, please review supported passwords."  

class OSExtensionNotFound(OSError):
    """ Raised when a searched extension is not found
    """ 

    def __str__(self) -> str:
        return f"Cannot locate extension. Please alter search criteria"
    
class OSRangeFault(OSError):
    """ Raised when a numeric range is invalid to the context provided, 
        E.G: uninitialised range values, disordered range values.
    """

    def __str__(self):
        return f"Range fault raised. Please verify integrity of passed range values"

class OSServiceNotAssigned(OSError):
    """ Raised a service needed is not assigned to a Broadworks entity.  
    """

    def __str__(self) -> str:
        return f"Service not assigend to target Broadworks entity. Please check services assigned."   

class OSFileNotFound(OSError):
    """ Raised when a file can not be found.  
    """

    def __str__(self) -> str:
        return f"File can not be found, please check path and file name."   
    
class OSApiResponseError(OSError):
    """ Raised when Odin Api returns an error code.
    """
    def __init__(self, response):
        response = response.json()
        self.response = f"{response['details']} {response['status']}: {response['error']}"

    def __str__(self) -> str:
        return self.response


import requests
import os

from .requester import Requester
from .methods import *
from .logger import Logger
from .scripter import Scripter
from .reporter import Reporter
from .exceptions import (OSApiAuthenticationFail, 
                                 OSSessionRefreshFail, 
                                 OSFailedToLocateSession)


class Api:
    
    def __init__(self, base_url: str, username: str, password: str,
                rate_limit: bool = True) -> None:
        """ Connection to Odin API, all interactions with the api are here.

        Args:
            base_url (str): Base url of your odin instance api.
            username (str): Username used when logging into odin account.
            password (str): Password used when logging into odin account stored as virtual environment.
            rate_limit (bool): Enables (True) or Disables (False) rate limiting to 5 calls per second. Defaults to True.
            
        Vars: 
            authorised (bool): Boolean value to indicate if api is authorised.\
            token (str): Token string returned from odin api.
        """
        
        self.base_url = base_url
        self.username = username
        self.password = os.getenv(password)
        self.rate_limit = rate_limit
        
        self.authorised = False
        self.token = ""
        
        self.logger = Logger.get_instance(self.username)
        self._requester = Requester(self.base_url, self.rate_limit, self.logger)
        
        self.get = get.Get(self._requester)
        self.post = post.Post(self._requester)
        self.put = put.Put(self._requester)
        self.delete = delete.Delete(self._requester)
        
        self.scripter = Scripter(api=self)
        self.reporter = Reporter(api=self)
 
 
    def authenticate(self):
        """Authenticates session with username and password supplied by user.

        Raises:
            OSApiAuthenticationFail: Raised if authenticaion fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """
        
        try:
            response = self.post.session(self.username, self.password)
            self._update_requester(response)
            return True
        except requests.exceptions.HTTPError:
            raise OSApiAuthenticationFail()
    
    
    def refresh_authorisation(self):
        """Re-authenticates the session with the API. Used if API key is to expire. 

        Raises:
            OSSessionRefreshFail: Raised if authentication fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """
        
        try:
            response = self.put.session()
            self._update_requester(response)
            return True
        except requests.exceptions.HTTPError:
            raise OSSessionRefreshFail()
  
    
    def get_auth_details(self):
        """Gets current session details. 

        Raises:
            OSFailedToLocateSession: Raised when session details can't be found.
                Most likely because session has expired.

        Returns:
            Dict: Current session details. 
        """
        
        try:
            return self.get.session()
        except requests.exceptions.HTTPError:
            raise OSFailedToLocateSession()


    def _update_requester(self, session_response):
        """When authenticating or re-auth update requester with token so it can make
        api calls 

        Args:
            session_response (obj): Requests mod response.
        """
        self.token = session_response["token"]
        self._requester.headers["Authorization"] = f"Bearer {self.token}"
        self.authorised = True


    def __str__(self) -> str:
        return f"API - url: {self.base_url}, username: {self.username}" \
            f"Authenticated: {self.authorised}"
    


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

setup(
    name="odins_spear",
    version="1.5.0",
    url="https://github.com/Jordan-Prescott/odins_spear",
    author="Jordan Prescott",
    author_email="jordan@minimalmind.io",
    description="Python library for Odin API to build and manage Broadworks installations.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=[
        "backports.tarfile==1.2.0",
        "certifi==2023.7.22",
        "cffi==1.16.0",
        "charset-normalizer==3.2.0",
        "colorama==0.4.6",
        "cryptography==41.0.7",
        "docutils==0.21.2",
        "et_xmlfile==2.0.0",
        "graphviz==0.20.1",
        "idna==3.4",
        "importlib_metadata==8.3.0",
        "jaraco.classes==3.4.0",
        "jaraco.context==6.0.1",
        "jaraco.functools==4.0.2",
        "keyring==25.3.0",
        "markdown-it-py==3.0.0",
        "mdurl==0.1.2",
        "more-itertools==10.4.0",
        "nh3==0.2.18",
        "numpy==1.26.4",
        "openpyxl==3.1.5",
        "pandas==2.2.2",
        "pkginfo==1.10.0",
        "pycparser==2.21",
        "Pygments==2.18.0",
        "PyJWT==2.8.0",
        "python-dateutil==2.9.0.post0",
        "pytz==2024.1",
        "pywin32-ctypes==0.2.3",
        "ratelimit==2.2.1",
        "readme_renderer==44.0",
        "requests==2.31.0",
        "requests-toolbelt==1.0.0",
        "rfc3986==2.0.0",
        "rich==13.7.1",
        "six==1.16.0",
        "tqdm==4.66.1",
        "twine==5.1.1",
        "tzdata==2024.1",
        "urllib3==2.0.4",
        "win32-setctime==1.1.0",
        "zipp==3.20.0",
    ],
)



__all__ = [
    "aa_cc_hg_audit",
    "bulk_enable_voicemail",
    "bulk_password_reset",
    "find_alias",
    "group_audit", 
    "move_numbers",
    "remove_numbers",
    "service_pack_audit",
    "service_provider_trunking_capacity",
    "user_activity",
    "user_association",
    "webex_builder"
]

from .aa_cc_hg_audit import main
from .bulk_enable_voicemail import main
from .bulk_password_reset import main
from .find_alias import main
from .group_audit import main
from .move_numbers import main
from .remove_numbers import main
from .service_pack_audit import main
from .service_provider_trunking_capacity import main
from .user_activity import main
from .user_association import main
from .user_registration import main
from .webex_builder import main
from .locate_free_extension import main



import requests
import json
from .exceptions import OSApiResponseError

from ratelimit import limits, sleep_and_retry
class Requester():

    def __init__(self, base_url: str, rate_limit: bool, logger: object = None):
        """Requester is the object that handles all request to and from the API. 
        Each method object will manage the setting up of the request however it is Requester
        that will send and receive the data. 

        Args:
            base_url (str): Base URL of API
            rate_limit (bool): Rate limit flag which will limit to 5 calls per 1 second
            logger (object, optional): Logger object for logging requests. Defaults to None.
        """
        
        self.base_url = base_url
        self.headers = {
            'Authorization': "",
            'Content-Type': 'application/json'
        }
        self.rate_limit = rate_limit
        self.logger = logger
    
    
    # get, post, put, delete methods takes in data and params and returns method type to _request
    def get(self, endpoint, data=None, params=None):
        return self._request(requests.get, endpoint, data, params)

    
    def post(self, endpoint, data=None):
        return self._request(requests.post, endpoint, data)

    
    def put(self, endpoint, data=None):
        return self._request(requests.put, endpoint, data)
    
    
    def delete(self, endpoint, data=None, params=None):
        return self._request(requests.delete, endpoint, data, params)


    def _request(self, method, endpoint, data=None, params=None):
        """Handles an API request with or without rate limiting.

        Args:
            method (request obj): Request module method type either GET, POST, PUT, DELETE
            endpoint (str): Specific API endpoint for functionality
            data (dict, optional): Python dict used in payload data if needed. Defaults to None.
            params (dict, optional): Parameters used in endpoint if needed. Defaults to None.

        Returns:
            API data: Data returned from API on completion of API call.
        """
        
        # if rate limiting is enabled uses _rate_limited_request where limiting is in place.
        if self.rate_limit:
            return self._rate_limited_request(method, endpoint, data, params)
        else:
            response = method(
                url=self.base_url + endpoint,
                headers=self.headers,
                data=json.dumps(data if data is not None else {}),
                params=(params if params is not None else {})
            )

            # if logger used log request
            if self.logger:
                self.logger._log_request(endpoint=endpoint, response_code=response.status_code)
                
            # flags errors if any returned from the API
            try:
                response.raise_for_status()
            except requests.exceptions.RequestException:
                raise OSApiResponseError(response)
            else:
                return response.json()
        

    @sleep_and_retry
    @limits(calls=5, period=1)
    def _rate_limited_request(self, method, endpoint, data=None, params=None):
        """Handles an API request with rate limiting.
        """
        
        response = method(
            url=self.base_url + endpoint,
            headers=self.headers,
            data=json.dumps(data if data is not None else {}),
            params=(params if params is not None else {})
        )
        
        # if logger used log request
        if self.logger:
            self.logger._log_request(endpoint=endpoint, response_code=response.status_code)
        
        # flags errors if any returned from the API
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException:
            raise OSApiResponseError(response)
        else:
            return response.json()
    


class Delete():
    def __init__(self, requester):
        self.requester = requester
    
#SESSION
#ACCOUNT AUTHORIZATION CODES
#ADMINISTRATORS
#ADVICE OF CHARGE
#ALTERNATE NUMBERS
#ANSWER CONFIRMATION
#ALTERNATE USER ID
#ANNOUNCEMENTS
#ANONYMOUS CALL REJECTION
#ATTENDANT CONSOLE
#AUTHENTICATION
#AUTO ATTENDANTS

    def auto_attendant(self, service_user_id: str):
        """Removes an Auto Attendant (AA) from a group. 

        Args:
            service_user_id(str): The service user ID of the AA.

        Returns:
            Dict: Returns the profile of the deleted AA.
        """

        endpoint = "/groups/auto-attendants"

        params = {
            "serviceUserId": service_user_id
        }

        return self.requester.delete(endpoint, params=params)


    def auto_attendant_submenu(self, service_user_id: str, submenu_id: str):
        """Removes an Auto Attendant (AA) Submenu from the AA configuration. Submenus are only a feature of the 'Auto Attendant - Standard' service. These are not available on Basic AAs.

        Args:
            service_user_id(str): The service user ID of the AA.
            submenu_id (str): The ID of the Submenu to be removed from the AA. 

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/auto-attendants/submenus"

        params = {
            "serviceUserId": service_user_id, 
            "submenuId": submenu_id
        }

        return self.requester.delete(endpoint, params=params)

#AUTOMATIC CALLBACK
#AUTOMATIC HOLD RETRIEVE
#BARGE IN EXEMPT
#BASIC CALL LOGS
#BROADWORKS ANYWHERE
#BROADWORKS MOBILITY
#BROADWORKS NAVIGATION
#BROADWORKS RECEPTIONIST ENTERPRISE
#BROADWORKS RECEPTIONIST OFFICE
#BROADWORKS RECEPTIONIST SMALL BUSINESS
#BUSY LAMP FIELD
#CALL CAPACITY
#CALL CENTER
#CALL CONTROL
#CALL FORWARDING ALWAYS
#CALL FORWARDING ALWAYS SECONDARY
#CALL FORWARDING BUSY
#CALL FORWARDING NO ANSWER
#CALL FORWARDING NOT REACHABLE
#CALL FORWARDING SELECTIVE
#CALL FORWARDING SETTINGS
#CALL NOTIFY
#CALL PARK
#CALL PICKUP
#CALL POLICIES
#CALL PROCESSING POLICIES
#CALL RECORDING
#CALL RECORDS
#CALL TRANSFER
#CALL WAITING
#CALLING LINE ID BLOCKING OVERRIDE
#CALLING LINE ID DELIVERY BLOCKING
#CALLING NAME DELIVERY
#CALLING NAME RETRIEVAL
#CALLING NUMBER DELIVERY
#CALLING PARTY CATEGORY
#CALLING PLANS
#CALLBACKS
#CHARGENUMBER
#CLASSMARK
#CLONE
#COLLABORATE
#COMM PILOT CALL MANAGER
#COMM PILOT EXPRESS
#COMMON PHONE LIST
#COMMUNICATION BARRING
#COMMUNICATION BARRING USER
#CONNECTED LINE IDENTIFICATION
#CUSTOM CONTACT DIRECTORY
#DEPARTMENTS
#DEVICE POLICIES
#DEVICES
#DIAL PLAN POLICY
#DIRECTED CALL PICKUP WITH BARGE IN
#DIRECTROUTE
#DN

    def group_dns(self, service_provider_id: str, group_id: str, 
                  start_of_range_number: str, end_of_range_number: str):
        """Removes range of numbers from a Group. 

        Args:
            service_provider_id (str): Service provider ID where the group is located. 
            group_id (str): Group ID where numbers are hosted.
            start_of_range_number (str): Starting number in range to remove from group. 
            end_of_range_number (str): Ending number in range to remove from group.

        Returns:
            None: This method does not return any specific value.
        """
        
        endpoint = "/groups/dns"
        
        data = {
			"serviceProviderId": service_provider_id,
			"groupId": group_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number,
                }
            ]
		}
        
        return self.requester.delete(endpoint, data=data)
    
    
    def service_provider_dns(self, service_provider_id: str, start_of_range_number: str, end_of_range_number: str):
        """Removes range of numbers from a Service Proiver or Enterprise. 

        Args:
            service_provider_id (str): Service provider ID where target numbers are located. 
            start_of_range_number (str): Starting number in range to remove from service provider. 
            end_of_range_number (str): Ending number in range to remove from service provider.

        Returns:
            None: This method does not return any specific value.
        """
        endpoint = "/service-providers/dns"
        
        data = {
			"serviceProviderId": service_provider_id,
			"dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number
                }
            ]
		}
        
        return self.requester.delete(endpoint, data=data)

#DO NOT DISTURB
#DOMAINS
#EMERGENCY NOTIFICATIONS
#EMERGENCY ZONES
#ENTERPRISE TRUNKS
#EXECUTIVE
#EXECUTIVE ASSISTANT
#EXTENSIONS
#EXTERNAL CALLING LINE ID DELIVERY
#EXTERNAL CUSTOM RINGBACK
#FAX MESSAGING
#FEATURE ACCESS CODES
#FLEXIBLE SEATING
#GROUP PAGING
#GROUPS
#GROUP NAVIGATION
#HOTELING GUEST
#HOTELING HOST
#HUNT GROUPS

    def group_hunt_group(self, service_user_id: str):

        """Deletes the specified hunt group.

        Args:
            service_user_id (str): The service user ID of the hunt group to be deleted.

        Returns:
            Dict: Profile of the deleted hunt group.

        """

        endpoint = "/groups/hunt-groups"

        params = {
            "serviceUserId": service_user_id
        }

        return self.requester.delete(endpoint, params=params)


#IN CALL SERVICE ACTIVATION
#INSTANT GROUP CALL
#INTEGRATED IMP
#INTERCEPT
#INTERNAL CALLING LINE ID DELIVERY
#LANGUAGES
#LEGACY AUTOMATIC CALLBACK
#MALICIOUS CALL TRACE
#MEDIA
#MEET-ME CONFERENCING
#MUSIC ON HOLD
#MWI DELIVERY TO MOBILE ENDPOINT
#NETWORK CLASS OF SERVICE
#NIGHT FORWARDING
#NUMBER PORTABILITY ANNOUNCEMENT
#NUMBER PORTABILITY QUERY
#NUMBERS
#OUTLOOK INTEGRATION
#PASSCODE RULES
#PASSWORD GENERATE
#PASSWORD RULES
#PERSONAL PHONE LIST
#PHONE DIRECTORY
#PHYSICAL LOCATION
#POLYCOM PHONE SERVICES
#PRE ALERTING
#PREFERRED CARRIER
#PREPAID
#PRIMARY ENDPOINT ADVANCED SETTING
#PRIORITY ALERT
#PRIVACY
#PUSH REGISTRATION
#PUSH TO TALK
#REGISTRATION
#REMOTE OFFICE
#REPORTS
#RESELLERS
#ROUTE LIST
#ROUTING PROFILE
#SCHEDULES
#SECURITY CLASSIFICATION
#SELECTIVE CALL ACCEPTANCE
#SELECTIVE CALL REJECTION
#SEQUENTIAL RING
#SERIES COMPLETION
#SERVICE PACKS
#SERVICE PROVIDERS
#SERVICES
#SHARED CALL APPEARANCE
#SILENT ALERTING
#SIMULTANEOUS RING PERSONAL
#SMARTY ADDRESS
#SPEED DIAL 100
#SPEED DIAL 8
#STATES AND PROVINCES
#SYSTEM
#TERMINATING ALTERNATE TRUNK IDENTITY
#THIRD PARTY VOICE MAIL SUPPORT
#THIRD PARTY EMERGENCY CALLING
#TIME ZONES
#TRUNK GROUPS
#TWO STAGE DIALING
#USERS
#USER CUSTOM RINGBACK
#VIDEO ADD ON
#VIRTUAL ON-NET ENTERPRISE EXTENSIONS
#VOICE MESSAGING
#VOICE PORTAL CALLING
#ZONE CALLING RESTRICTIONS
#ODIN BRANDING
#ODIN CALLBACKS
#ODIN CONNECTORS
#ODIN EMAIL
#ODIN EVENTS
#ODIN INVENTORY
#ODIN REPORTS
#ODIN SETTINGS
#ODIN STATUS
#ODIN SSO
#ODIN SSO ALTERNATE USER IDS
#ODIN TASKS
#ODIN UI
#ODIN VIEWABLE PACKS
#ODIN WEBHOOKS
#ODIN AUDIT
#ODIN IMPORTS
#ODIN EXPORTS
#XSI
#SIP AUTHENTICATION
#POLICY
#PASSWORD RESET
#PARTNERS
#USER UTILITIES
#ODIN TASKS COPY
#ODIN CONNECTORS
#LOCALES




import re 

def find_entity_with_number_type(number: str, number_type: str, broadwork_entities: list):
    """
    Finds the phone number, extension, or alias in list of broadwork entities 
    and returns that entity once found. 

    Args:
        number (str): Target number looking for e.g. +1-123456789 or alias 0
        number_type (str): Number type of either phone number, extenion, or alias
        broadwork_entities (list): List of broadwork entities number can be assigned to

    Returns:
        object: Broadwork entity where number is assigned.
    """
    
    for entity in broadwork_entities:
        try:
            if number_type == 'dn' and number in entity.phone_number:
                return entity
            elif number_type == 'extension' and number in entity.extension:
                return entity
            elif number_type == 'alias':
                for alias in entity.aliases:
                    if re.search(rf'\b{re.escape(alias)}\b', number):
                        return entity
        except TypeError:
            # issue when entity does not have a number, extenion, or alias assigned
            continue
    return None


from ..exceptions import OSExtensionNotFound, OSRangeFault

def retrieve_extensions( api, service_provider_id: str, group_id: str ) -> list:
    
    extensions = []

    dataset = (
       api.get.users( service_provider_id, group_id              ) +
       api.get.group_hunt_groups( service_provider_id, group_id  ) +
       api.get.group_call_centers( service_provider_id, group_id ) +
       api.get.auto_attendants( service_provider_id, group_id    )
    )

    for data in dataset:
        if not data['extension']:
            continue

        extensions.append( int(data['extension']) )

    return extensions if extensions else None

def main( api, service_provider_id: str, group_id: str, range_start: int, range_end: int ):
    '''Retrieves The Lowest Free Extension Available In The Designated Group Passed.
    '''

    if range_start > range_end:
        raise OSRangeFault

    # Retrieve List Of Occupied Extensions Within The Group
    extensions = retrieve_extensions(
        api,
        service_provider_id,
        group_id,
    )

    for extension in range(range_start, range_end + 1):
        if extension not in extensions:
            return {'extension': extension}

    raise OSExtensionNotFound